--------------------------------------------------------------------------------
-- Title:               LUMap.lua
-- Description:         Like a square peg in a round hole
-- Author:              Raphaël Szwarc http://alt.textdrive.com/lua/
-- Creation Date:       February 1, 2005
-- Legal:               Copyright (C) 2005 Raphaël Szwarc
--------------------------------------------------------------------------------

-- import dependencies
require( "LUObject" )
require( "LUList" )

-- parent class
local superClass = LUObject

-- inherit the super class methods
local thisClass = superClass.inherit( superClass )

-- factory class method, this is where all the instance methods are implemented as closures
thisClass.new = function()
        -- instantiate the super class
        local super = superClass.new()
        
        -- inherit the super class instance methods
        local this = super.class().inherit( super )
        
        -- local table for the instance variables
        local ivars = {}

        -- private instance method to convert a given key to a string
        local stringForKey = function( aKey )
                if aKey ~= nil and type( aKey ) ~= "string" then
                        if this.isKindOf( aKey ) == true then
                                aKey = aKey.toString()
                        else
                                aKey = tostring( aKey )
                        end
                end
                
                return aKey
        end
        
        -- instance method to check if this map contains a given key
        this.hasKey = function( aKey )
                if aKey ~= nil then
                        aKey = stringForKey( aKey )
                
                        if ivars[ aKey ] then
                                return true
                        end
                end
                
                return false
        end

        -- instance method to check if this map contains a given value
        this.hasValue = function( aValue )
                if aValue ~= nil then
                        local isKindOf = this.isKindOf( aValue )
                        
                        for aKey, anotherValue in ivars do
                                if aValue == anotherValue then
                                        return true
                                end
                        
                                if isKindOf == true and aValue.equals( anotherValue ) == true then
                                        return true
                                end
                        end
                end
                
                return false
        end

        -- instance method to return a value for a given key
        this.get = function( aKey )
                if aKey ~= nil then
                        aKey = stringForKey( aKey )
                        
                        return ivars[ aKey ]
                end
                
                return nil
        end

        -- instance method to compare a given object with this map for equality
        this.equals = function( anObject )
                if this == anObject then
                        return true
                end
                
                if this.isKindOf( anObject ) == true then
                        if this.size() == anObject.size() then
                                for aKey, aValue in ivars do
                                        local anotherValue = anObject.get( aKey )
                                        
                                        if anotherValue ~= nil then
                                                if this.isKindOf( anotherValue ) == true and aValue.equals( anotherValue ) == false then
                                                        return false
                                                end
                                                
                                                if aValue ~= anotherValue then
                                                        return false
                                                end
                                        else
                                                return false
                                        end
                                end
                                
                                return true
                        end
                end
                
                return false
        end

        -- instance method to access all the keys of this map
        this.keys = function()
                local someKeys = LUList.new()
                
                for aKey, aValue in ivars do
                        someKeys.add( aKey )
                end
                
                return someKeys
        end

        -- instance method to associate a given value with a given key
        this.put = function( aKey, aValue )
                if aKey ~= nil then
                        aKey = stringForKey( aKey )
                
                        ivars[ aKey ] = aValue
                end

                return this
        end

        -- instance method to copies all the key-value pairs from a given map
        this.putAll = function( aMap )
                if this.isKindOf( aMap ) then
                        local   someKeys = aMap.keys()
                        local   count = someKeys.size()
                        
                        for index = 1, count do
                                local aKey = someKeys.get( index )
                                local aValue = aMap.get( aKey )
                                
                                this.put( aKey, aValue )
                        end
                elseif type( aMap ) == 'table' then
                        for aKey, aValue in aMap do
                                this.put( aKey, aValue )
                        end
                end
                
                return this
        end
        
        -- instance method to returns the size of this map
        this.size = function()
                local aSize = 0
                
                for aKey, aValue in ivars do
                        aSize = aSize + 1
                end
                
                return aSize
        end

        -- instance method for string representation
        this.toString = function()
                local aTable = {}
                
                table.insert( aTable, "{ " )
                
                if this.size() > 0 then
                        for aKey, aValue in ivars do
                                local aKeyDescription = nil
                                local aValueDescription = nil
                                
                                if this.isKindOf( aKey ) == true then
                                        aKeyDescription = aKey.toString()
                                else
                                        aKeyDescription = tostring( aKey )
                                end

                                if this.isKindOf( aValue ) == true then
                                        aValueDescription = aValue.toString()
                                else
                                        aValueDescription = tostring( aValue )
                                end

                                table.insert( aTable, string.format( "%q", aKeyDescription ) )
                                table.insert( aTable, " = " )

                                table.insert( aTable, string.format( "%q", aValueDescription ) )
                                
                                table.insert( aTable, "; " )
                        end

                        table.remove( aTable, table.getn( aTable ) )
                end
                
                table.insert( aTable, " }" )
                
                return table.concat( aTable, "" )
        end

        -- instance method to access all the values of this map
        this.values = function()
                local someValues = LUList.new()
                
                for aKey, aValue in ivars do
                        someValues.add( aValue )
                end
                
                return someValues
        end

        -- return the public instance methods
        return this
end

-- define the class name
LUMap = thisClass

return LUMap